/* -*- Mode: C++; indent-tabs-mode: nil; c-basic-offset: 2; tab-width: 2 -*- */
/***************************************************************************
 *            Thread.h
 *
 *  FUPPES - Free UPnP Entertainment Service
 *
 *  Copyright (C) 2009-2011 Ulrich Völkel <u-voelkel@users.sourceforge.net>
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef _THREAD_H
#define _THREAD_H

#ifdef HAVE_CONFIG_H
#include "../../config.h"
#endif

#ifdef WIN32
//#pragma comment(lib,"shlwapi.lib")
#include <shlwapi.h>
#else
#include <pthread.h>
#include <unistd.h>
#endif

#include <string>
#include <map>
#include <list>

namespace fuppes {


#ifdef WIN32
  typedef DWORD       threadId_t;
#else
  typedef pthread_t   threadId_t;
#endif
  
  
class Mutex
{
	public:
		Mutex();
		~Mutex();

		void lock();
		void unlock();

		bool locked() const { return m_locked; }
		
	private:
		#ifdef WIN32
		CRITICAL_SECTION  m_mutex;
		#else
		pthread_mutex_t   m_mutex;
		#endif
		bool m_locked;
};


class MutexLocker
{
	public:
		MutexLocker(Mutex* mutex) {
			m_mutex = mutex;
			m_mutex->lock();
		}
		~MutexLocker() {
			m_mutex->unlock();
		}

	private:
		Mutex* m_mutex;
};


class Thread
{
	public:
		Thread(std::string name);
		virtual ~Thread();

    bool start(void* arg = NULL);
		void stop() { m_stop = true; }
    bool stopRequested() const { return m_stop; };
		bool close();

		bool running() const { return m_running; }
		bool finished() const { return m_finished; }

		void name(const std::string name) { m_name = name; }
		std::string name() const { return m_name; }

#ifdef WIN32
    threadId_t threadId() { return m_threadId; }    
    static inline threadId_t currentThreadId() { return GetCurrentThreadId(); }
#else
    threadId_t threadId() { return m_handle; }
    static inline threadId_t currentThreadId() { return pthread_self(); }
#endif
		
  protected:
		virtual void run() = 0;
		
		#ifdef WIN32
    static DWORD threadFunc(void*);
    #else
    static void* threadFunc(void*);
    #endif

    void msleep(unsigned int milliseconds);    
		
  private:
    
		std::string m_name;
    void* m_arg;
		bool	m_running;
		bool  m_finished;
		bool  m_stop;
    
	#ifdef WIN32
		HANDLE		      m_handle;
    DWORD           m_threadId;
	#else
		pthread_t				m_handle;
		pthread_cond_t	m_exitCondition;
		pthread_mutex_t m_mutex;
	#endif
    
};


class MainThread: public Thread
{
  public:
    MainThread(threadId_t id):Thread("main") {
      m_threadId = id;
    }

    threadId_t threadId() { return m_threadId; }
    
  private:
    void run() { }
    threadId_t  m_threadId;
};

class ThreadPool: public Thread
{
  public:
    static void init();
    static void uninit();

    static void add(Thread* thread);
    static void remove(Thread* thread);

    static Thread* threadById(threadId_t id);

    static void deleteLater(Thread* thread);

    static void lock() { return m_instance->m_mutex.lock(); }
    static std::map<threadId_t, Thread*>* threads() { return &m_instance->m_threads; }
    static void unlock() { return m_instance->m_mutex.unlock(); }
    
  private:
    ThreadPool();
    ~ThreadPool();

    void run();
    
    static ThreadPool* m_instance;

    Mutex   m_mutex;

    Thread*                            m_mainThread;
    threadId_t                         m_mainThreadId;
    std::map<threadId_t, Thread*>      m_threads;

    std::list<Thread*>                 m_deleteLater;
};

}

#endif // _THREAD_H
