/* -*- Mode: C++; indent-tabs-mode: nil; c-basic-offset: 2; tab-width: 2 -*- */
/***************************************************************************
 *            MacAddressTable.cpp
 *
 *  FUPPES - Free UPnP Entertainment Service
 *
 *  Copyright (C) 2009 Ulrich Völkel <u-voelkel@users.sourceforge.net>
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "MacAddressTable.h"
#include "../Common/File.h"
#include "../Common/RegEx.h"

#ifdef WIN32
#include <stdio.h>
#include <iphlpapi.h>
#endif

using namespace fuppes;

#include <iostream>
using namespace std;

#define MAC_UNKNOWN "00:00:00:00:00:00"

MacAddressTable* MacAddressTable::m_instance = NULL;

void MacAddressTable::init() // static
{
  if(m_instance == NULL)
    m_instance = new MacAddressTable();
}

MacAddressTable* MacAddressTable::instance() // static
{
  init();
  return m_instance;
}

void MacAddressTable::uninit() // static
{
  if(m_instance == NULL)
    return;

  clear();
  delete m_instance;
}


bool MacAddressTable::mac(std::string ip, std::string& mac) // static
{
  MutexLocker locker(&instance()->m_mutex);

  std::map<std::string, std::string>::iterator it;
  it = instance()->m_map.find(ip);
  if(it != instance()->m_map.end()) {
    mac = it->second;
    return true;
  }

  mac = instance()->getMac(ip);
  if(mac != MAC_UNKNOWN) {
    instance()->m_map[ip] = mac;
    return true;
  }
  return false;
}

void MacAddressTable::clear() // static
{
  MutexLocker locker(&instance()->m_mutex);
  instance()->m_map.clear();
}


std::string MacAddressTable::getMac(std::string ip) // static
{
#ifdef WIN32
  PMIB_IPNETTABLE pIpNetTable = NULL;
  ULONG dwSize = 0;
  DWORD ret;

  // get the required size and alloc the buffer
  ret = GetIpNetTable(NULL, &dwSize, false);
  pIpNetTable = (MIB_IPNETTABLE*)malloc(dwSize);
  // get the table
  ret = GetIpNetTable(pIpNetTable, &dwSize, false);

  // successfull but no dat
  if(ret == ERROR_NO_DATA) {
    free(pIpNetTable);
    return MAC_UNKNOWN;
  }

  // error
  if(ret != NO_ERROR) {
    std::cout << "MacAddressTable::getMac: error" << std::endl;
    if(ret == ERROR_INSUFFICIENT_BUFFER)
      std::cout << "ERROR_INSUFFICIENT_BUFFER" << std::endl;
    else if(ret == ERROR_INVALID_PARAMETER)
      std::cout << "ERROR_INVALID_PARAMETER" << std::endl;

    free(pIpNetTable);
    return MAC_UNKNOWN;
  }

  for(DWORD i = 0; i < pIpNetTable->dwNumEntries; i++) {

    std::string addr = inet_ntoa(*(struct in_addr *)&pIpNetTable->table[i].dwAddr);
    if(addr != ip || pIpNetTable->table[i].dwPhysAddrLen != 6)
      continue;

    char mac[18];
    memset(mac, 0, sizeof(mac));

    sprintf(mac,"%.2x:%.2x:%.2x:%.2x:%.2x:%.2x",
	    pIpNetTable->table[i].bPhysAddr[0],
		  pIpNetTable->table[i].bPhysAddr[1],
      pIpNetTable->table[i].bPhysAddr[2],
      pIpNetTable->table[i].bPhysAddr[3],
      pIpNetTable->table[i].bPhysAddr[4],
      pIpNetTable->table[i].bPhysAddr[5]);

    free(pIpNetTable);
    return mac;
  }

  free(pIpNetTable);
  return MAC_UNKNOWN;

#else

  File file("/proc/net/arp");
  if(file.open(File::Read)) {
    std::string line;
    RegEx rx("([\\d|\\.]+) +0x\\d +0x\\d +([\\d|\\w|:]+)");
    while(file.getline(line)) {

      if(!rx.search(line))
        continue;

      if(rx.match(1) == ip) {
        file.close();
        return rx.match(2);
      }
    }
    file.close();
  }

  return MAC_UNKNOWN;
#endif  
}

